﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;

namespace howto_spiralize
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private Color[] Colors =
        {
            Color.Red, Color.OrangeRed, Color.Yellow,
            Color.Green, Color.Blue, Color.Indigo,
            Color.Fuchsia,
        };
        private Bitmap OriginalBm = null;

        // Draw the spiralized image.
        private void DrawImage()
        {
            if (OriginalBm == null) return;
            float A, color_scale, opacity;
            if (!float.TryParse(txtA.Text, out A)) return;
            if (!float.TryParse(txtColorScale.Text, out color_scale)) return;
            if (!float.TryParse(txtOpacity.Text, out opacity)) return;

            picSpiral.Image = SpiralizeImage(
                OriginalBm, Colors, A, color_scale, opacity,
                chkOutlineSpirals.Checked, chkElliptical.Checked);
        }

        // Spiralize the image.
        private Bitmap SpiralizeImage(Bitmap original_bm, Color[] colors,
            float A, float color_scale, float opacity,
            bool outline_spirals, bool make_elliptical)
        {
            int wid = original_bm.Width;
            int hgt = original_bm.Height;
            Bitmap bm = new Bitmap(wid, hgt);
            using (Graphics gr = Graphics.FromImage(bm))
            {
                gr.Clear(picSpiral.BackColor);
                gr.SmoothingMode = SmoothingMode.AntiAlias;

                int num_spirals = colors.Length;

                // Angular spacing between different spirals.
                float d_start = (float)(2 * Math.PI / num_spirals);

                // The angle where the next spiral starts.
                float start_angle = 0;

                // Center point.
                PointF center = new PointF(wid / 2, hgt / 2);

                // Draw on the entire image.
                Rectangle rect = new Rectangle(0, 0, wid, hgt);

                // Find the maximum distance to the rectangle's corners.
                float max_dist = Distance(center, rect);

                // Calculate the maximum theta value that we need to go to.
                float max_theta = max_dist / A + 2 * (float)Math.PI;

                // Get points defining the spirals.
                List<List<PointF>> spiral_points = new List<List<PointF>>();

                // Get the spirals' points.
                for (int i = 0; i <= num_spirals; i++)
                {
                    spiral_points.Add(GetSpiralPoints(
                        center, A, start_angle, max_theta));
                    start_angle += d_start;
                }

                // Fill the areas between the spirals.
                for (int i = 0; i < num_spirals; i++)
                {
                    // Make a list holding the next spiral's points.
                    List<PointF> points = new List<PointF>(spiral_points[i]);

                    // Add the following spiral's points reversed.
                    List<PointF> points2 =
                        new List<PointF>(spiral_points[i + 1]);
                    points2.Reverse();
                    points.AddRange(points2);

                    // Make a colorized version of the image.
                    using (Bitmap colorized_bm =
                        ColorizeImage(OriginalBm, Colors[i], color_scale))
                    {
                        // Fill the spiral.
                        using (TextureBrush brush =
                            new TextureBrush(colorized_bm))
                        {
                            gr.FillPolygon(brush, points.ToArray());
                        }
                    }

                    // Optional: Outline the spiral's polygon.
                    if (outline_spirals)
                        gr.DrawLines(Pens.Black, points.ToArray());
                }
            }

            // Combine the new image with the original one.
            Bitmap new_bm = (Bitmap)original_bm.Clone();
            using (Graphics gr = Graphics.FromImage(new_bm))
            {
                // Make a ColorMatrix with the opacity.
                ColorMatrix color_matrix = new ColorMatrix();
                color_matrix.Matrix33 = opacity;

                // Make the ImageAttributes object.
                ImageAttributes attributes = new ImageAttributes();
                attributes.SetColorMatrix(color_matrix,
                    ColorMatrixFlag.Default, ColorAdjustType.Bitmap);

                // Draw the input bitmap onto the Graphics object.
                Rectangle rect = new Rectangle(0, 0,
                    new_bm.Width, new_bm.Height);

                gr.DrawImage(bm, rect,
                    0, 0, new_bm.Width, new_bm.Height,
                    GraphicsUnit.Pixel, attributes);
            }

            // Make it elliptical if desired.
            if (make_elliptical) new_bm = MakeElliptical(new_bm);

            return new_bm;
        }

        // Return points that define a spiral.
        // See http://csharphelper.com/blog/2018/10/draw-a-filled-spiral-in-c/
        private List<PointF> GetSpiralPoints(
            PointF center, float A,
            float angle_offset, float max_theta)
        {
            // Get the points.
            List<PointF> points = new List<PointF>();
            const float dtheta = (float)(5 * Math.PI / 180);    // Five degrees.
            for (float theta = 0; ; theta += dtheta)
            {
                // Calculate r.
                float r = A * theta;

                // Convert to Cartesian coordinates.
                float x, y;
                PolarToCartesian(r, theta + angle_offset, out x, out y);

                // Center.
                x += center.X;
                y += center.Y;

                // Create the point.
                points.Add(new PointF((float)x, (float)y));

                // If we have gone far enough, stop.
                if (theta + angle_offset > max_theta) break;
            }
            return points;
        }

        // Return the distance between two points.
        private float Distance(PointF point1, PointF point2)
        {
            float dx = point1.X - point2.X;
            float dy = point1.Y - point2.Y;
            return (float)Math.Sqrt(dx * dx + dy * dy);
        }

        // Return the distance between the point
        // and the rectangle's farthest corner.
        private float Distance(PointF point, Rectangle rect)
        {
            float max_dist = Distance(point, new PointF(rect.Left, rect.Top));

            float test_dist = Distance(point, new PointF(rect.Left, rect.Bottom));
            if (max_dist < test_dist) max_dist = test_dist;

            test_dist = Distance(point, new PointF(rect.Right, rect.Top));
            if (max_dist < test_dist) max_dist = test_dist;

            test_dist = Distance(point, new PointF(rect.Right, rect.Bottom));
            if (max_dist < test_dist) max_dist = test_dist;

            return max_dist;
        }

        // Convert polar coordinates into Cartesian coordinates.
        private void PolarToCartesian(float r, float theta, out float x, out float y)
        {
            x = (float)(r * Math.Cos(theta));
            y = (float)(r * Math.Sin(theta));
        }

        // Return a colorized image.
        // See http://csharphelper.com/blog/2014/10/rainbowize-an-image-in-c/
        private Bitmap ColorizeImage(Bitmap original_bm,
            Color color, float color_scale)
        {
            // Create a ColorMatrix for this color.
            ColorMatrix cm = new ColorMatrix(new float[][]
            {
                new float[] {color.R / 255f * color_scale, 0, 0, 0, 0},
                new float[] {0, color.G / 255f * color_scale, 0, 0, 0},
                new float[] {0, 0, color.B / 255f * color_scale, 0, 0},
                new float[] {0, 0, 0, 1, 0},
                new float[] {0, 0, 0, 0, 1},
            });

            // Make the result bitmap.
            int wid = original_bm.Width;
            int hgt = original_bm.Height;
            Bitmap result_bm = new Bitmap(wid, hgt);

            // Color the image.
            using (ImageAttributes attr = new ImageAttributes())
            {
                attr.SetColorMatrix(cm);

                // Copy the original image onto the result bitmap.
                using (Graphics gr = Graphics.FromImage(result_bm))
                {
                    Point[] dest_points =
                {
                    new Point(0, 0),
                    new Point(wid, 0),
                    new Point(0, hgt),
                };
                    Rectangle source_rect = new Rectangle(0, 0, wid, hgt);
                    gr.DrawImage(original_bm, dest_points,
                        source_rect, GraphicsUnit.Pixel, attr);
                }
            }

            return result_bm;
        }

        private void mnuFileOpen_Click(object sender, EventArgs e)
        {
            if (ofdFile.ShowDialog() == DialogResult.OK)
            {
                OriginalBm = LoadBitmapUnlocked(ofdFile.FileName);
                DrawImage();
            }
        }

        private void mnuFileSaveAs_Click(object sender, EventArgs e)
        {
            if (sfdFile.ShowDialog() == DialogResult.OK)
            {
                SaveImage(picSpiral.Image, sfdFile.FileName);
            }
        }

        private void txtParameter_TextChanged(object sender, EventArgs e)
        {
            DrawImage();
        }

        // Save the file with the appropriate format.
        public void SaveImage(Image image, string filename)
        {
            string extension = Path.GetExtension(filename);
            switch (extension.ToLower())
            {
                case ".bmp":
                    image.Save(filename, ImageFormat.Bmp);
                    break;
                case ".exif":
                    image.Save(filename, ImageFormat.Exif);
                    break;
                case ".gif":
                    image.Save(filename, ImageFormat.Gif);
                    break;
                case ".jpg":
                case ".jpeg":
                    image.Save(filename, ImageFormat.Jpeg);
                    break;
                case ".png":
                    image.Save(filename, ImageFormat.Png);
                    break;
                case ".tif":
                case ".tiff":
                    image.Save(filename, ImageFormat.Tiff);
                    break;
                default:
                    throw new NotSupportedException(
                        "Unknown file extension " + extension);
            }
        }

        // Load a bitmap without locking it.
        private Bitmap LoadBitmapUnlocked(string file_name)
        {
            using (Bitmap bm = new Bitmap(file_name))
            {
                return new Bitmap(bm);
            }
        }

        private void chkParameter_CheckedChanged(object sender, EventArgs e)
        {
            DrawImage();
        }

        // Make the image elliptical.
        private Bitmap MakeElliptical(Bitmap bm)
        {
            Bitmap new_bm = (Bitmap)bm.Clone();
            using (Graphics gr = Graphics.FromImage(new_bm))
            {
                gr.Clear(Color.Transparent);
                using (TextureBrush brush = new TextureBrush(bm))
                {
                    Rectangle rect = new Rectangle(0, 0, bm.Width, bm.Height);
                    gr.FillEllipse(brush, rect);
                }
            }
            return new_bm;
        }
    }
}
